/************* REGULAR CONTEXT SEQUENCE MATCHING  ***********************

File: rwcs.pl
Author: Temur Kutsia
Date: 26.03.2006
Last modification: 01.10.2008
Implementation of regular well-moded context sequence constraint solving.
Copyright (c) 2006-2008 Temur Kutsia

This program will run on the following Prolog dialects:
SWI-Prolog, GNU-Prolog, BinProlog, LPA Win-Prolog. It does not run
on YAP Prolog.

=========================================================================

Input syntax:

  Context variables:
      Identifiers that start with lower case letters c_
      and contain at least three letters. 
      Anonymous context variables are written as c_.
      Examples: c_1, c_C, c_ThisIsCtxVar.

  Sequence variables:
      Identifiers that start with lower case letters s_.
      and contain at least three letters. 
      Anonymous sequence variables are written as s_.
      Examples: s_1, s_S, s_ThisIsSeqVar.
 
  Function variables:
      Identifiers that start with lower case letters f_.
      and contain at least three letters.
      Anonymous function variables are written as f_.
      Examples: f_1, f_F, f_ThisIsFunVar.
  
  Individual variables:
      Identifiers that start with lower case letters i_.
      and contain at least three letters.
      Anonymous function variables are written as i_.
      Examples: i_1, i_X, i_ThisIsIndVar.

  Constants:
      Prolog atoms that do not start with c_, s_, f_, i_.
      Examples: See any textbook on Prolog.

  Reserved words:
      hole, eps, app. THEY SHOULD NOT BE USED IN THE INPUT PROBLEM.
      sstar, sor, sconc, cstar, cor, cconc. THEY CAN BE USED ONLY AS REGULAR OPERATORS.

  Terms: Built using the grammar:
      t ::= iv | sv | hole | f(t1,...,tn) | fv(t1,...,tn) | cv(t)
      where f is a flexible arity function constant, and t in cv(t)
      is not a sequence variable.
      NOTE: Current implementation does not check input for syntactical
      correctness. It is solely user's responcibility to provide a
      syntactically correct input. In the next versions it will be added.

  Contexts: Terms with only one hole.

  Equation:
      E ::= t1 << t2 
      where t1 and t2 are not sequence variables and do not contain hole.

  Well-Moded Equational Constraint:
      WMEC ::= [t11 << t12,..., tn1 << tn2]
      such that t12 is ground and variables of each ti2 (1 < i =< n)
      occur in tj1, j < i.

  Regular Expressions:
      Regular expressions for hole-free terms or Regular expressions for contexts:
      E ::= T | C

  Regular Expressions for hole-free terms:
      Built using the grammar:
      T ::= i_ | sv | f(T) | fv(T) | cv(T) | eps | sconc(T1,T2) |
        sor(T1,T2) | sstar(T) | C(T)
      where head of T in cv(T) and C(T) is either an individual variable,
      function variable, context variable, or function symbol.

      If C is a context (i.e. a special case of regular expressions
      for contexts without regular operators) then in C(T) the T is
      inserted instead of the hole in C. Otherwise C(T) is written as
      app(C,T).

      For h being a function symbol or function variable:
        h(E1, sconc(T1, T2), E2) is identified to h(E1, T1, T2, E2).
        h(E1, eps, E2) is identified to h(E1, E2).

  Regular Expressions for Contexts:
      Built using the grammar:
      C ::= hole | f(T1, C, T2) | fv(T1, C, T2) | cv(C) | cconc(C1,C2) | 
            cor(C1,C2) | cstar(C)
      

  Membership atoms:
      MA ::= [sv, T, fl] | [cv, C, fl] 
      where fl (called a flag) is an expression that evaluates to 0 or 1.
      The user can also write [sv, T] for [sv, T, 0] and [cv, C] for [cv, C, 0].

  Well-moded membership constraints:
      WMMC ::= [[v1, E1, fl1],...,[vn, En, fln]]
      where n>=0, v's are context or sequence variables, Ei is T if vi is a
      sequence variable and is C otherwise, fl's are flags, and for each i
      (1 =< i =< n) vi does not occur in Ej, j >= i.

  Regular well-moded context sequence constraints:
      RWCS ::= [E1,...,En, MC1,...,MCm]
      where [E1,...,En] is WMEC, [MC1,...,MCm] is WMMC, and  if
      for some i and j (i \= j) MCi = [v, Ei, fli]  and MCj = [v, Ej, flj]
      then v iccurs in some Ek (1 >= k >= n).



Output syntax:

  Bindings:
      B ::= iv ---> t | sv ---> (t1,...,tn) | fv ---> f | cv ---> C
      where n>=0, f is a function symbol and C is a context.
      iv \= t, sv \= (t1,...,tn), cv(hole) \= C.
  Substitutions:
      S ::= [B1,...,Bn]
      n>=0.
       
=========================================================================

References to the Regular Well-moded Context Sequence Constraint Solving:

Temur Kutsia and Mircea Marin. 
A Rule-Based Framework for Solving Regular Context Sequence Constraints. 
Technical Report 06-01. 
RISC, Johannes Kepler University, Linz, Austria, 2006.

************************************************************************/

:-dynamic(numb/1).

:-op(120,xfy,--->).
%:-op(110,yf,*).



                /**********************************
                 *    RWCS: The main predicate    *
                 **********************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
rwcs(Constraint, Solution) accepts a regular well-moded context sequence
constraint Constraint. If Constraint is solvable, a solving substitution 
is returned in Solution. The user can get all the answers one by one, 
by pressing the ';' key after each answer.

rwcs works as follows: First, the terms in the input problem are converted
into lists of symbols. Then a solution is computed. Next, if the solution
contains anonymous variables it is recomputed after instantiating anonymous
variables with fresh singleton variables. From the obtained substitution all
bindings for temporary variables (i.e., those variables that do not occur in
Constraint) are removed. Finally, the output is translated into term
substituions.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */



rwcs(Constraint,Solution):-
   term_constraints_to_symbol_list_constraints(Constraint,SLConstr),
   solve(SLConstr,SmallerSWithoutAnonymous),
   symbol_list_substitutions_to_term_substitutions(SmallerSWithoutAnonymous,
                                                   Solution).


solve(SLConstr,SmallerSWithoutAnonymous):-                                                   
   asserta(numb(1000)),
   match(SLConstr,[],S),
   recompute_match_with_anonymous_variables_instantiated(SLConstr,S,
                                                         SWithoutAnonymous),
   remove_bindings_for_temporary_variables(SWithoutAnonymous,
                                           SmallerSWithoutAnonymous),
   retractall(numb).                                             


             /*******************************************************
             *    Conversions from input format to internal form    *
             *******************************************************/

% term_to_list transforms a term into its list representation.
% Example: the term f_F(a,c_G(s_X,h(b)),i_Y) will be translated into
% [f_F, a, [s_G, s_X, [h, b]], i_Y]

term_to_list(Var,Term):-var(Var),
        term_to_atom(Var,Term),
        !.

term_to_list(Ctx_or_seq_or_fun_or_ind_var_or_constant,X):-
        functor(Ctx_or_seq_or_fun_or_ind_var_or_constant,_,0),
        !,
        transform_var_or_constant(Ctx_or_seq_or_fun_or_ind_var_or_constant,X).

term_to_list(sstar(T),sstar(List)):-!,
    term_to_list(T,List).

term_to_list(sconc(T1,T2),sconc(List1,List2)):-!,
    term_to_list(T1,List1),
    term_to_list(T2,List2).

term_to_list(sor(T1,T2),sor(List1,List2)):-!,
    term_to_list(T1,List1),
    term_to_list(T2,List2).

term_to_list(eps,eps):-!.

term_to_list(cstar(T),cstar(List)):-!,
    term_to_list(T,List).

term_to_list(cconc(T1,T2),cconc(List1,List2)):-!,
    term_to_list(T1,List1),
    term_to_list(T2,List2).

term_to_list(cor(T1,T2),cor(List1,List2)):-!,
    term_to_list(T1,List1),
    term_to_list(T2,List2).
       

term_to_list(Term,List):-Term=..PseudoList,
        pseudo_list_to_list(PseudoList,List).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Transforms variables into terms of the form ctx_var(Var), seq_var(Var),
 fun_var(Var), or ind_var(Var). Leaves constants unchanged.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


transform_var_or_constant(Ind_Var,ind_var(Ind_Var)):-
        ind_var(Ind_Var),
        !.
transform_var_or_constant(Fun_Var,fun_var(Fun_Var)):-
        fun_var(Fun_Var),
        !.
transform_var_or_constant(Seq_Var,seq_var(Seq_Var)):-
        seq_var(Seq_Var),
        !.
transform_var_or_constant(Ctx_Var,ctx_var(Ctx_Var)):-
        ctx_var(Ctx_Var),
        !.
transform_var_or_constant(Term,Term):-functor(Term,_,0),
        !.

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Checks for nonanonymous context, sequence, function and individual
 variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ind_var([_|_]):-!,fail.
ind_var(X):-name(X,[105,95,_|_]).

fun_var([_|_]):-!,fail.
fun_var(X):-name(X,[102,95,_|_]).

seq_var([_|_]):-!,fail.
seq_var(X):-name(X,[115,95,_|_]).

ctx_var([_|_]):-!,fail.
ctx_var(X):-name(X,[99,95,_|_]).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 pseudo_list_to_list takes a list of symbols and terms and returns
 a list of symbols and list representations of terms.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

pseudo_list_to_list([],[]).
pseudo_list_to_list([Term|OtherTerms],[TermList|OtherTermsList]):-
       term_to_list(Term,TermList0),
       flatten_comma(TermList0,TermList),
       pseudo_list_to_list(OtherTerms,OtherTermsList).

flatten_comma([',',[','|T1],[','|T2]],Res):-
           !,
           append(T1,T2,R),
           flatten_comma([','|R],Res).

flatten_comma([',',[','|T1],T2],Res):-
           !,
           append(T1,T2,R),
           flatten_comma([','|R],Res).

flatten_comma([',',T1,[','|T2]],[',',T1|R]):-
           !,
           flatten_comma([','|T2],[','|R]).

flatten_comma([',',T1,T2],[',',TR1,TR2]):-
           !,
           flatten_comma(T1,TR1),
           flatten_comma(T2,TR2).

flatten_comma([],[]):-!.
flatten_comma([X|L],[X1|L1]):- !,
           flatten_comma(X,X1),
           flatten_comma(L,L1).

flatten_comma(X,X).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 term_constraints_to_symbol_list_constraints transforms
 terms in constraints into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

term_constraints_to_symbol_list_constraints([],[]).
term_constraints_to_symbol_list_constraints([S << T|Rest],
                        [SL << TL|RestL]):-
        term_to_list(S,SL),
        term_to_list(T,TL),
        term_constraints_to_symbol_list_constraints(Rest,RestL).

term_constraints_to_symbol_list_constraints([[Var,Regexp,IntegerFlag]|Rest],
                        [[seq_var(Var),RegexpL,
                        IntegerFlag]|RestL]):-
    seq_var(Var),!,
        regexp_to_list(Regexp,RegexpL),
        term_constraints_to_symbol_list_constraints(Rest,RestL).

term_constraints_to_symbol_list_constraints([[Var,Regexp,IntegerFlag]|Rest],
                        [[ctx_var(Var),RegexpL,
                        IntegerFlag]|RestL]):-
    ctx_var(Var),!,
        regexp_to_list(Regexp,RegexpL),
        term_constraints_to_symbol_list_constraints(Rest,RestL).

term_constraints_to_symbol_list_constraints([[Var,Regexp]|Rest],L):-
        term_constraints_to_symbol_list_constraints([[Var,Regexp,0]|Rest],L).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 regexp_to_list transforms regular expressions
 into their symbol list representations
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*  Uncomment this if your Prolog dialect supports postfix * and infix | and ','.
regexp_to_list(R*,star(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list((R1|R2),or(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list((R1,R2),conc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
*/
regexp_to_list(sstar(R),sstar(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list(sor(R1,R2),sor(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(sconc(R1,R2),sconc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(eps,eps):-!.
regexp_to_list(cstar(R),cstar(RL)):-!,
        regexp_to_list(R,RL).
regexp_to_list(cor(R1,R2),cor(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(cconc(R1,R2),cconc(R1L,R2L)):-!,
        regexp_to_list(R1,R1L),
        regexp_to_list(R2,R2L).
regexp_to_list(Expr,ExprList):-
        term_to_list(Expr,ExprList).


        /*******************************************************
        *  Conversions from internal form to the input format  *
        ********************************************************/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 symbol_list_substitutions_to_term_substitutions transforms
 symbol list representations of terms in substitutions
 into their input format
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

symbol_list_substitutions_to_term_substitutions([],[]).
symbol_list_substitutions_to_term_substitutions([X--->TL |Rest1],
                                                [Y---> T|Rest2]):-
        symbol_list_to_term(X,Y),
        symbol_list_to_term(TL,T),
        symbol_list_substitutions_to_term_substitutions(Rest1,Rest2).


symbol_list_to_term(ctx_var(Var),Var):-!.
symbol_list_to_term(fun_var(Var),Var):-!.
symbol_list_to_term(seq_var(Var),Var):-!.
symbol_list_to_term(ind_var(Var),Var):-!.
symbol_list_to_term(eps,eps):-!.
symbol_list_to_term(seq([X1]),X2):-!,
        symbol_list_to_term(X1,X2).
symbol_list_to_term(seq([H1,H2|Tail]),(X,Y)):-!,
        symbol_list_to_term(H1,X),
        symbol_list_to_term(seq([H2|Tail]),Y).
symbol_list_to_term([Head|Tail],Term):-!,
        list_of_symbol_lists_to_symbol_list(Tail,T1),
        Term=..[Head|T1].
symbol_list_to_term(Term,Term).


list_of_symbol_lists_to_symbol_list([],[]).
list_of_symbol_lists_to_symbol_list([X|T],[X|T1]):-
        functor(X,_,0),
        !,
        list_of_symbol_lists_to_symbol_list(T,T1).
list_of_symbol_lists_to_symbol_list([X|T],[X1|T1]):-
        symbol_list_to_term(X,X1),
        list_of_symbol_lists_to_symbol_list(T,T1).


        /*******************************************************
        *   Get substituitions that bind input variables only    *
        ********************************************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_bindings_for_temporary_variables gets a substitution as the input
and removes from it the bindings for all those variables that did not occur
in the input problem but have been generated at run-time.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

remove_bindings_for_temporary_variables([],[]).
remove_bindings_for_temporary_variables([ctx_var(Var) ---> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([fun_var(Var) ---> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([seq_var(Var) ---> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([ind_var(Var) ---> _ |Tail],
                                        NewSubst):-
        name(Var,List),
        append(_,[116, 101, 109, 112|_],List),
        !,
        remove_bindings_for_temporary_variables(Tail, NewSubst).
remove_bindings_for_temporary_variables([Var ---> T |Tail],[Var ---> T |Tail1]):-
        remove_bindings_for_temporary_variables(Tail,Tail1).
        


        /*******************************
        *    Transformation rules      *
        *******************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Basic transformation rules
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Success
match([],Acc,Acc).

% Trivial
match([X << X|Tail],Acc,Subst):-!,match(Tail,Acc,Subst).

% Nullary funtions

match([[X|Xs] << X|Tail],Acc,Subst):-
        !,
        match([[X|Xs] << [X]|Tail],Acc,Subst).
match([X << [X]|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular constraints for terms
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% IRET  (Intersection of regular expressions for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
    append(L1,[[seq_var(SeqVar),T1,Flag1]|L2],Tail),
        append(L2_1,[[seq_var(SeqVar),T2,Flag2]|L2_2],L2),
    !,
        create_unique_temporary_seqvar(seq_var(SeqVar),TempSeqVar),
        append(L1,[[seq_var(SeqVar),T1,Flag1]|L2_1],L3),
        append(L3,[[TempSeqVar,T2,Flag2]|L2_2],NewTail),
        match([[X,seq_var(SeqVar)|Xs] << [X|Ys],
               [X,TempSeqVar] << [X, seq_var(SeqVar)]|NewTail],
          Acc,Subst).

% VA  (Variable abstraction in regular expressions for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
    append(L1,[[seq_var(SeqVar),T,Flag]|L2],Tail),
        purify(T,PureT,NewConstraints),
        !,
        append(L1,[[seq_var(SeqVar),PureT,Flag]|L2],L3),
        append(L3,NewConstraints,NewTail),
        match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|NewTail],Acc,Subst).


% ESRET (Empty sequence in a regular expression for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),eps,Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[seq_var(SeqVar) ---> eps],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) ---> eps],NewXs),
        apply_substitution(L3,[seq_var(SeqVar) ---> eps],NewTail),
        match([[X|NewXs] << [X|Ys]|NewTail],NewAcc,Subst).

% CRET (Concatenation in a regular expression for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),sconc(R1,R2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_seq_var(Flag,Value),
        create_unique_temporary_seqvar(seq_var(SeqVar),
                                              TempSeqVar1),
        create_unique_temporary_seqvar(seq_var(SeqVar),
                                              TempSeqVar2),
        new_flags_for_concatenation(Value,TempSeqVar1,NewFlag1,NewFlag2),
        compose(Acc,[seq_var(SeqVar) --->
                     seq([TempSeqVar1,TempSeqVar2])],NewAcc),
        instance_of_term_list(Xs,
                              [seq_var(SeqVar) --->
                               seq([TempSeqVar1,TempSeqVar2])],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) --->
                               seq([TempSeqVar1,TempSeqVar2])],Tail1),
        NewTail=[[TempSeqVar1,R1,NewFlag1],[TempSeqVar2,R2,NewFlag2]|Tail1],
        match([[X,TempSeqVar1,TempSeqVar2|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% ChRET (Choice in a regular expression for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),sor(R1,R2),Flag]|L2],Tail),
        !,
        select_branch(sor(R1,R2),Rselected),
        evaluate_flag_for_seq_var(Flag,Value),
        append([[seq_var(SeqVar),Rselected,Value]|L1],L2,NewTail),
        match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|NewTail],Acc,Subst).

% RRET1 (Repetition in a regular expression for terms)


match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),sstar(_),Flag]|L2],Tail),
        evaluate_flag_for_seq_var(Flag,0),
        compose(Acc,[seq_var(SeqVar) ---> eps],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) ---> eps],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) ---> eps],NewTail),
        match([[X|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

% RRET2 (Repetition in a regular expression for terms)


match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),sstar(T),_]|L2],Tail),
        !,
        create_unique_temporary_seqvar(seq_var(SeqVar),TempSeqVar),
        compose(Acc,[seq_var(SeqVar) --->
                     seq([TempSeqVar,seq_var(SeqVar)])],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) --->
                                  seq([TempSeqVar,seq_var(SeqVar)])],
                              Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) --->
                               seq([TempSeqVar,seq_var(SeqVar)])],
                           Tail1),
        NewTail=[[TempSeqVar,T,1],[seq_var(SeqVar),sstar(T),0]|Tail1],
        match([[X,TempSeqVar,seq_var(SeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

% SVRET (Sequence variable in a regular expression for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),seq_var(AnotherSeqVar),0]|L2],Tail),
        !,
        compose(Acc,[seq_var(SeqVar) ---> seq_var(AnotherSeqVar)],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) --->
                                  seq_var(AnotherSeqVar)],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) --->
                               seq_var(AnotherSeqVar)],NewTail),
        match([[X,seq_var(AnotherSeqVar)|Xs1] << [X|Ys]|NewTail],
              NewAcc,Subst).

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),
                    seq_var(AnotherSeqVar),1]|L2],Tail),
        !,
        create_unique_temporary_indvar(ind_var(AnotherSeqVar),
                       ind_var(FreshIndVar)),
        create_unique_temporary_seqvar(seq_var(AnotherSeqVar),
                       seq_var(FreshSeqVar)),
        compose(Acc,[seq_var(SeqVar) ---> seq_var(AnotherSeqVar)],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) --->
                                  seq_var(AnotherSeqVar)],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) --->
                               seq_var(AnotherSeqVar)],NewTail),
        match([[X,seq_var(AnotherSeqVar)|Xs1] << [X|Ys], 
               [X,ind_var(FreshIndVar),seq_var(FreshSeqVar)] << 
               [X,seq_var(AnotherSeqVar)]|NewTail],NewAcc,Subst).


% TRET (term in a regular expression for terms)

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        append(L1,[[seq_var(SeqVar),Term,_]|L2],Tail),
        !,
        compose(Acc,[seq_var(SeqVar) ---> Term],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) ---> Term],Xs1),
        append(L1,L2,L3),
        apply_substitution(L3,[seq_var(SeqVar) ---> Term],NewTail),
        match([[X,Term|Xs1] << [X|Ys]|NewTail],NewAcc,Subst).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Transformation rules for regular expressions on contexts
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% IREC  (Intersection of regular expressions for contexts)

match([[ctx_var(CtxVar),Arg] << T | Tail],Acc,Subst):-
    append(L1,[[ctx_var(CtxVar),T1,Flag1]|L2],Tail),
        append(L2_1,[[ctx_var(CtxVar),T2,Flag2]|L2_2],L2),
    !,
        create_unique_temporary_ctxvar(ctx_var(CtxVar),TempCtxVar),
        append(L1,[[ctx_var(CtxVar),T1,Flag1]|L2_1],L3),
        append(L3,[[TempCtxVar,T2,Flag2]|L2_2],NewTail),
    create_unique_temporary_constant(Const),
        match([[ctx_var(CtxVar),Arg] << T,
               [TempCtxVar,Const] << [ctx_var(CtxVar),Const]|NewTail],
          Acc,Subst).

% VAC  (Variable abstraction in regular expressions for terms)

match([[ctx_var(CtxVar),Arg] << T |Tail],Acc,Subst):-
    append(L1,[[ctx_var(CtxVar),C,Flag]|L2],Tail),
        purify(C,PureC,NewConstraints),
        !,
        append(L1,[[ctx_var(CtxVar),PureC,Flag]|L2],L3),
        append(L3,NewConstraints,NewTail),
        match([[ctx_var(CtxVar),Arg] << T |NewTail],Acc,Subst).


% HREC (Hole in a regular expression for contexts)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),hole,Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,0),
        append(L1,L2,L3),
        compose(Acc,[ctx_var(CtxVar) ---> hole],NewAcc),
        instance(Arg,[ctx_var(CtxVar) ---> hole],NewArg),
        apply_substitution(L3,[ctx_var(CtxVar) ---> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% CREC (Concatenation in a regular expression for contexts)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),cconc(C1,C2),Flag]|L2],Tail),
        !,
        evaluate_flag_for_ctx_var(Flag,Value),
        create_unique_temporary_ctxvar(ctx_var(CtxVar),TempCtxVar1),
        create_unique_temporary_ctxvar(ctx_var(CtxVar),TempCtxVar2),
        new_flags_for_concatenation(Value,TempCtxVar1,NewFlag1,NewFlag2),
        compose(Acc,[ctx_var(CtxVar)--->[TempCtxVar1,[TempCtxVar2,hole]]],
                NewAcc),
        instance(Arg,[ctx_var(CtxVar)--->[TempCtxVar1,[TempCtxVar2,hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar)--->
                               [TempCtxVar1,[TempCtxVar2,hole]]],
                           Tail1),
        NewTail=[[TempCtxVar1,C1,NewFlag1],[TempCtxVar2,C2,NewFlag2]|Tail1],
        match([[TempCtxVar1,[TempCtxVar2,NewArg]] << T|NewTail],NewAcc,Subst).

% ChREC (Choice in a regular expression for contexts)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),cor(C1,C2),Flag]|L2],Tail),
        !,
        select_branch(cor(C1,C2),Rselected),
        evaluate_flag_for_ctx_var(Flag,Value),
        append([[ctx_var(CtxVar),Rselected,Value]|L1],L2,NewTail),
        match([[ctx_var(CtxVar),Arg] << T|NewTail],Acc,Subst).

% RREC1 (Repetition in a regular expression for contexts)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),cstar(_),Flag]|L2],Tail),
        evaluate_flag_for_ctx_var(Flag,0),
        compose(Acc,[ctx_var(CtxVar) ---> hole],NewAcc),
        instance(Arg,[ctx_var(CtxVar) ---> hole],NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar) ---> hole],NewTail),
        match([NewArg << T|NewTail],NewAcc,Subst).

% RREC2 (Repetition in a regular expression for contexts)


match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),cstar(R),_]|L2],Tail),
        !,
        create_unique_temporary_ctxvar(ctx_var(CtxVar),TempCtxVar),
        compose(Acc,[ctx_var(CtxVar) --->
                     [TempCtxVar,[ctx_var(CtxVar),hole]]],
                NewAcc),
        instance(Arg,[ctx_var(CtxVar) --->
                      [TempCtxVar,[ctx_var(CtxVar),hole]]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar) --->
                               [TempCtxVar,[ctx_var(CtxVar),hole]]],
                           Tail1),
        NewTail=[[TempCtxVar,R,1],[ctx_var(CtxVar),cstar(R),0]|Tail1],
        match([[TempCtxVar,[ctx_var(CtxVar),NewArg]] << T|NewTail],
              NewAcc,Subst).

% CVREC (Context variable in a regular expression for contexts)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),
            [ctx_var(AnotherCtxVar),hole],0]|L2],
               Tail),
        !,
        compose(Acc,[ctx_var(CtxVar)--->[ctx_var(AnotherCtxVar),hole]],
                NewAcc),
        instance(Arg,[ctx_var(CtxVar)--->[ctx_var(AnotherCtxVar),hole]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar) --->
                               [ctx_var(AnotherCtxVar),hole]],
                           NewTail),
        match([[ctx_var(AnotherCtxVar),NewArg] << T|NewTail],
              NewAcc,Subst).

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),
                   [ctx_var(AnotherCtxVar),hole],1]|L2],
               Tail),
        !,
        create_unique_temporary_funvar(fun_var(AnotherCtxVar),
                       fun_var(FunVar)),
        create_unique_temporary_seqvar(seq_var(AnotherCtxVar),
                       seq_var(SeqVar1)),
        create_unique_temporary_ctxvar(ctx_var(AnotherCtxVar),
                       ctx_var(FreshCtxVar)),
        create_unique_temporary_seqvar(seq_var(AnotherCtxVar),
                       seq_var(SeqVar2)),
    create_unique_temporary_constant(Const),
    compose(Acc,[ctx_var(CtxVar)--->[ctx_var(AnotherCtxVar),hole]],
                NewAcc),
        instance(Arg,[ctx_var(CtxVar)--->[ctx_var(AnotherCtxVar),hole]],
                 NewArg),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar) --->
                               [ctx_var(AnotherCtxVar),hole]],
                           NewTail),
    match([[ctx_var(AnotherCtxVar),NewArg] << T,
           [fun_var(FunVar),seq_var(SeqVar1),
        [ctx_var(FreshCtxVar),Const],
        seq_var(SeqVar2)] << [ctx_var(AnotherCtxVar),Const]|NewTail],
              NewAcc,Subst).

% CxRET1 (Context in a regular expression for contexts 1)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),
            [ctx_var(AnotherCtxVar),C],Flag]|L2],Tail),
        !,
    append(L1,[[ctx_var(CtxVar),
            cconc([ctx_var(AnotherCtxVar),hole],C),Flag]|L2],NewTail),
    match([[ctx_var(CtxVar),Arg] << T|NewTail],Acc,Subst).

% CxRET2 (Context in a regular expression for contexts 2)

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        append(L1,[[ctx_var(CtxVar),Ctx,_]|L2],Tail),
%        context(Ctx),
        !,
        compose(Acc,[ctx_var(CtxVar) ---> Ctx],NewAcc),
        apply_substitution([[ctx_var(CtxVar),Arg] << T],
                           [ctx_var(CtxVar) ---> Ctx],NewEquation),
        append(L1,L2,L3),
        apply_substitution(L3,[ctx_var(CtxVar) ---> Ctx],NewTail),
        append(NewEquation,NewTail,NewProblem),
        match(NewProblem,NewAcc,Subst).
       

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Rules that do not operate on regular expressions
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


% Sequence variable deletion

match([[X,s_|Xs] << [X|Ys]|Tail],Acc,Subst):-
        match([[X|Xs] << [X|Ys]|Tail],Acc,Subst).

match([[X,seq_var(SeqVar)|Xs] << [X|Ys]|Tail],Acc,Subst):-
        compose(Acc,[seq_var(SeqVar) ---> eps],NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) ---> eps],Xs1),
        apply_substitution(Tail,[seq_var(SeqVar) ---> eps],Tail1),
        match([[X|Xs1] << [X|Ys]|Tail1],NewAcc,Subst).

% Widening

match([[X,s_|Xs] << [X,_|Ys]|Tail],Acc,Subst):-
        !,
        match([[X,s_|Xs] << [X|Ys]|Tail],Acc,Subst).


match([[X,seq_var(SeqVar)|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[seq_var(SeqVar) --->
                     seq([T1,seq_var(SeqVar)])],
                NewAcc),
        instance_of_term_list(Xs,[seq_var(SeqVar) --->
                                  seq([T1,seq_var(SeqVar)])],
                              Xs1),
        apply_substitution(Tail,[seq_var(SeqVar) --->
                                 seq([T1,seq_var(SeqVar)])],
                           Tail1),
        match([[X,seq_var(SeqVar)|Xs1] << [X|Ys]|Tail1],
              NewAcc,Subst).

% Context variable deletion

match([[c_,Arg] << T|Tail],Acc,Subst):-
        match([Arg << T|Tail],Acc,Subst).

match([[ctx_var(CtxVar),Arg] << T|Tail],Acc,Subst):-
        \+(member([ctx_var(CtxVar),_RegExp,_Flag],Tail)),
        compose(Acc,[ctx_var(CtxVar) ---> hole],NewAcc),
        instance(Arg,[ctx_var(CtxVar) ---> hole],Arg1),
        apply_substitution(Tail,[ctx_var(CtxVar) ---> hole],Tail1),
        match([Arg1 << T|Tail1],NewAcc,Subst).

% Deepening

match([[c_,Arg] << [F|Args]|Tail],Acc,Subst):-
        append([F|_],[Proj|_],[F|Args]),
        match([[c_,Arg] << Proj|Tail],Acc,Subst).

match([[ctx_var(CtxVar),Arg] << [F|Args]|Tail],Acc,Subst):-
        \+(member([ctx_var(CtxVar),_RegExp,_Flag],Tail)),
        append([F|L1],[Proj|L2],[F|Args]),
        append([F|L1],[[ctx_var(CtxVar),hole]|L2],Context),
        compose(Acc,[ctx_var(CtxVar) ---> Context],NewAcc),
        instance(Arg,[ctx_var(CtxVar) ---> Context],Arg1),
        apply_substitution(Tail,[ctx_var(CtxVar) ---> Context],Tail1),
        match([[ctx_var(CtxVar),Arg1] << Proj|Tail1],NewAcc,Subst).

% Decomposition

match([[X,S1|Xs] << [X,T1|Ys]|Tail],Acc,Subst):-!,
        match([S1 << T1,[X|Xs] << [X|Ys]|Tail],Acc,Subst).

% Function variable elimination

match([[f_|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        match([[Y|Xs] << [Y|Ys]|Tail],Acc,Subst).

match([[fun_var(X)|Xs] << [Y|Ys]|Tail],Acc,Subst):-
        !,
        compose(Acc,[fun_var(X) ---> Y],NewAcc),
        instance_of_term_list(Xs,[fun_var(X) ---> Y],Xs1),
        apply_substitution(Tail,[fun_var(X) ---> Y],Tail1),
        match([[Y|Xs1] << [Y|Ys]|Tail1],NewAcc,Subst).

match([[f_|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        match([[T|Xs] << [T]|Tail],Acc,Subst).
        
match([f_ << T|Tail],Acc,Subst):-
        match([[f_] << T|Tail],Acc,Subst).
        
match([[fun_var(X)|Xs] << T|Tail],Acc,Subst):-
        T\=[_|_],
        T\=[],
        !,
        compose(Acc,[fun_var(X) ---> T],NewAcc),
        instance_of_term_list(Xs,[fun_var(X) ---> T],Xs1),
        apply_substitution(Tail,[fun_var(X) ---> T],Tail1),
        match([[T|Xs1] << [T]|Tail1],NewAcc,Subst).
        
match([fun_var(X) << T|Tail],Acc,Subst):-
        match([[fun_var(X)] << T|Tail],Acc,Subst).
        

        

% Individual variable elimination

match([i_ << _Term|Tail],Acc,Subst):-
        !,
        match(Tail,Acc,Subst).

match([ind_var(X) << Y|Tail],Acc,Subst):-
        !,
        compose(Acc,[ind_var(X) ---> Y],NewAcc),
        apply_substitution(Tail,[ind_var(X) ---> Y],Tail1),
        match(Tail1,NewAcc,Subst).

match([[X,_,_]|Tail],_,_):-
      \+ member(_ << _, Tail),
      member([X,_,_],Tail),
      (X=seq_var(Y); X=ctx_var(Y)),
      write('Error. Variable '), 
      write(Y),
      write(' constrained by several constraints does not occur in equations.'),
      nl,abort.

% SAT

match([[_,_,0]|Tail],Acc,Subst):-
      \+ member(_ << _, Tail),
      match(Tail,Acc,Subst).

match([[X,_,1]|Tail],_,_):-
      \+ member(_ << _, Tail),
      (X=seq_var(Y); X=ctx_var(Y)),
      write('Error. Constrained variable '),
      write(Y),
      write(', whose flag is 1, does not occur in equations.'),
      nl,abort.


      /*******************************************************************
      *    Recomputing solutions with instantiated anonymous variables   *
      *******************************************************************/

recompute_match_with_anonymous_variables_instantiated(MP,
                                                      OldMatcher,
                                                      NewMatcher):-
        anonymous_to_singleton_variables_in_subst(OldMatcher,
                                                  MatcherWithoutAnonymous),
        recompute_matchers(MP,OldMatcher,MatcherWithoutAnonymous,NewMatcher).


recompute_matchers(_,MatcherWithoutAnonymous,
                   MatcherWithoutAnonymous,MatcherWithoutAnonymous):-!.
recompute_matchers(MP,_,MatcherWithoutAnonymous,NewMatcher):-
        select_equations(MP,Eq),
        apply_substitution(Eq, MatcherWithoutAnonymous,NewMP),
        match(NewMP, MatcherWithoutAnonymous,NewMatcher).


        /*******************************
        *    Check for contexts.       *
        *******************************/


% context([hole|Tail]):-!,without_holes(Tail).
% context([H|Tail]):-context(H),!,without_holes(Tail).
% context([H|Tail]):-without_holes(H),!,context(Tail).

% Check whether a term is without holes, or whether terms in a list
% are all without holes.

% without_holes([]):-!.
% without_holes([X|Tail]):-without_holes(X),without_holes(Tail).
% without_holes(X):-X\=hole.



        /*********************************************
        *        Operations on substitutions         *
        *********************************************/


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
compose(S1,S2,S) composes two substitutions S1 and S2 and returns S.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

compose(S1,S2,S):-
        apply_substitution(S1,S2,S3),
        remove_duplicated_vars_and_append_substs(S3,S2,S).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
apply_substitution takes as input arguments the lists S1 and S2, and returns
S3. S1 may contain bindings, matching equations, regular constraints, and
terms. S2 is a substitution. S3 is obtained from S1 by applying S2 to
terms in it (except variables X in regular constraints [X,R,F]).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

apply_substitution([],_,[]):-!.
apply_substitution(S,[],S):-!.
apply_substitution([X ---> Y | Tail],S,[X ---> Y1 | Tail1]):-!,
        instance(Y,S,Y1),
        apply_substitution(Tail, S, Tail1).
apply_substitution([X << Y | Tail],S,[X1 << Y1 | Tail1]):- !,
        instance(X,S,X1),
    instance(Y,S,Y1),
        apply_substitution(Tail,S,Tail1).
apply_substitution([[X,Regexp,Flag]|Tail],S,[[X,Regexp1,Flag1]|Tail1]):-!,
        instance(Regexp,S,Regexp1),
        instance(Flag,S,Flag1),
        apply_substitution(Tail,S,Tail1).
apply_substitution([X|Tail],S,[X1 |Tail1]):-
        instance(X,S,X1),
        apply_substitution(Tail,S,Tail1).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
instance(X,Subst,I) returns an instance I of X with respect to the
substitution Subst.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
instance(circplus(X,Y),Subst,circplus(XInst,YInst)):-!,
        instance(X,Subst,XInst),
        instance(Y,Subst,YInst). */
instance(trivial(X),Subst,trivial(XInst)):-!,
        instance(X,Subst,XInst).
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_sequence(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_regexp(TermSequence_or_Regexp_or_Term,Subst,Instance),
        !.
instance(TermSequence_or_Regexp_or_Term,Subst,Instance):-
        instance_of_nonseqvar_term(TermSequence_or_Regexp_or_Term,Subst,
                                   Instance).


% Instance of seq expressions.

instance_of_sequence(seq(L),Subst,seq(Instance)):-
        instance_of_term_list(L,Subst,LInst),
        flatten_sequence(LInst,Instance).
        
instance_of_sequence(seq_var(SeqVar),Subst,Instance):-
        !,
        instance_of_symbol(seq_var(SeqVar),Subst,Instance).


% Instance of regular expressions.

instance_of_regexp(cstar(C),Subst,cstar(Instance)):-!,
        instance_of_regexp(C,Subst,Instance).
instance_of_regexp(cor(C1,C2),Subst,cor(R1Inst,R2Inst)):-!,
        instance_of_regexp(C1,Subst,R1Inst),
        instance_of_regexp(C2,Subst,R2Inst).
instance_of_regexp(cconc(C1,C2),Subst,cconc(R1Inst,R2Inst)):-!,
        instance_of_regexp(C1,Subst,R1Inst),
        instance_of_regexp(C2,Subst,R2Inst).
instance_of_regexp(sstar(T),Subst,sstar(Instance)):-!,
        instance_of_regexp(T,Subst,Instance).
instance_of_regexp(sor(T1,T2),Subst,sor(R1Inst,R2Inst)):-!,
        instance_of_regexp(T1,Subst,R1Inst),
        instance_of_regexp(T2,Subst,R2Inst).
instance_of_regexp(sconc(T1,T2),Subst,sconc(R1Inst,R2Inst)):-!,
        instance_of_regexp(T1,Subst,R1Inst),
        instance_of_regexp(T2,Subst,R2Inst).
instance_of_regexp(eps,_,eps):-!.
instance_of_regexp(hole,_,hole):-!.
instance_of_regexp([ctx_var(CtxVar),hole],Subst,Context):-!,
        instance_of_ctx_var(ctx_var(CtxVar),Subst,Context).
instance_of_regexp(seq_var(SeqVar),Subst,Conc):-
 %       SeqVar\=[_|_],        % what is this?
        !,
        instance_of_symbol(seq_var(SeqVar),Subst,Sequence),
        seq_to_conc(Sequence,Conc).
instance_of_regexp(Term,Subst,Instance):-
        instance_of_nonseqvar_term(Term,Subst,Instance).


% Instance of terms that are not sequence variables

instance_of_nonseqvar_term([ctx_var(CtxVar),Arg],Subst,Instance):-
        !,
        instance_of_ctx_var(ctx_var(CtxVar),Subst,Context),
        instance_of_nonseqvar_term(Arg,Subst,ArgInst),
        apply_context(Context,ArgInst,Instance).

instance_of_nonseqvar_term([FunVar_or_FunSymb|ArgList],Subst,
                           [FunSymb|ArgInst]):-
        !,
        instance_of_symbol(FunVar_or_FunSymb,Subst,FunSymb),
        instance_of_term_list(ArgList,Subst,ArgInst).

instance_of_nonseqvar_term(Symbol,Subst,Instance):-
        instance_of_symbol(Symbol,Subst,Instance).


% Instance of a list of terms.

instance_of_term_list([seq_var(SeqVar)|OtherTerms],Subst,Instance):-
 %       SeqVar \= [_|_],   % do we still need it?
        !,
        instance_of_symbol(seq_var(SeqVar),Subst,Sequence),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst),
        flatten_sequence([Sequence|OtherTermsInst],Instance).

instance_of_term_list([Term|OtherTerms],Subst,
                      [TermInstance|OtherTermsInst]):-
        instance_of_nonseqvar_term(Term,Subst,TermInstance),
        instance_of_term_list(OtherTerms,Subst,OtherTermsInst).

instance_of_term_list([],_,[]).


% Instance of a context variable.

instance_of_ctx_var(ctx_var(CtxVar),Subst,Instance):-
        member(ctx_var(CtxVar)--->Instance,Subst),
        !.
instance_of_ctx_var(ctx_var(CtxVar),_,[ctx_var(CtxVar),hole]).


% Instance of a symbol.

instance_of_symbol(Var,Subst,Instance):-
        member(Var--->Instance,Subst),
        !.
instance_of_symbol(Constant_or_var_not_in_dom,_,
                   Constant_or_var_not_in_dom).


% Transforming seq expressions into concatenation.

seq_to_conc(seq([X]),X):-!.
seq_to_conc(seq([X,Y|T]),sconc(X,R)):-seq_to_conc(seq([Y|T]),R).
seq_to_conc(eps,eps).


% Context application.

apply_context(hole,Term,Term):-!.
apply_context([hole],Term,Term):-!.
apply_context([Term1,hole],Term,[Term1,Term]):-!.
apply_context([hole|CtxTail],Term,[Term|CtxTail]):-!.
apply_context([[CxtHead|CtxTail]|Tail],Term,[NewTerm|Tail]):-
        apply_context([CxtHead|CtxTail],Term,NewTerm).
apply_context([Head|Tail],Term,[Head|NewTail]):-
        apply_context(Tail,Term,NewTail).

% Flattening nested seq expressions.

flatten_sequence([eps|Tail],Tail):-!.
flatten_sequence([seq([])|Tail],Tail):-!.
flatten_sequence([seq([X|T])|Tail],[X|Flat]):-!,
        flatten_sequence([seq(T)|Tail],Flat).
flatten_sequence([X|Tail],[X1|Flat]):-!,
        flatten_sequence(X,X1),
        flatten_sequence(Tail,Flat).
flatten_sequence(X,X).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
remove_duplicated_vars_and_append_substs(S1,S2,S) takes two lists of bindings
S1 and S2, deletes all the bindings X--->T from S2 such that S1 contains a
binding for X, deletes all the bindings of the form X--->X from S1, and
joins the obtained lists together into S.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


remove_duplicated_vars_and_append_substs([],L,L):-!.
remove_duplicated_vars_and_append_substs(L,[],L):-!.
remove_duplicated_vars_and_append_substs(Subst1,Subst2,Result):-
        remove_vars_from_second_that_occur_in_first(Subst1,Subst2,
                                                    [],Subst2New),
        remove_identities(Subst1,Subst1New),
        append(Subst1New,Subst2New,Result).
remove_vars_from_second_that_occur_in_first([X--->_|Rest1],[X--->_|Rest2],
                                            ResultSoFar,Result):-
        !,
        remove_vars_from_second_that_occur_in_first(Rest1,Rest2,
                                                    ResultSoFar,Result).
remove_vars_from_second_that_occur_in_first([X--->Y|Rest1],[B|Rest2],
                                            Cleaned,Result):-
        !,
        remove_vars_from_second_that_occur_in_first([X--->Y|Rest1],Rest2,
                                                    [B|Cleaned],Result).
remove_vars_from_second_that_occur_in_first([_|Rest1],[],CleanedOfFirstVar,
                                            Result):-!,
        remove_vars_from_second_that_occur_in_first(Rest1,CleanedOfFirstVar,
                                                    [],Result).
remove_vars_from_second_that_occur_in_first([],Result,_,Result).

remove_identities([],[]).
remove_identities([X--->X|Rest],Result):-
        !,
        remove_identities(Rest,Result).
remove_identities([B|Rest],[B|Result]):-
        remove_identities(Rest,Result).



                 /*******************************
                 *      Various utilities       *
                 *******************************/



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Generating unique temporary context, sequence, function and individual
variables.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

create_unique_temporary_ctxvar(ctx_var(CtxVar),
                   ctx_var(TempCtxVar)):-
    create_unique_temporary_symbol(CtxVar,TempCtxVar).

create_unique_temporary_seqvar(seq_var(SeqVar),
                   seq_var(TempSeqVar)):-
    create_unique_temporary_symbol(SeqVar,TempSeqVar).

create_unique_temporary_funvar(fun_var(FunVar),
                   fun_var(TempFunVar)):-
    create_unique_temporary_symbol(FunVar,TempFunVar).

create_unique_temporary_indvar(ind_var(IndVar),
                   ind_var(TempIndVar)):-
    create_unique_temporary_symbol(IndVar,TempIndVar).

create_unique_temporary_constant(Const):-
    create_unique_temporary_symbol(const,Const).

create_unique_temporary_symbol(Input,Unique):-
        name(Input,L),
    name(temp, TempL),
        append(L,TempL,TempNameBaseList),
        name(TempNameBase,TempNameBaseList),
        my_gensym(TempNameBase, Unique).

my_gensym(Base,Unique):-
        numb(N),
        name(Base,L),
        name(N,NL),
        append(L,NL,L1),
        name(Unique,L1),
        N1 is N + 1,
        retract(numb(N)),
        asserta(numb(N1)),!.


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Flag manipulation in regular constraints: Generating new flags for the
constraints that arise after splitting concatenation, and evaluating
flag values.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

% New flag generation
new_flags_for_concatenation(0,_,0,0).
new_flags_for_concatenation(1,Var,0,trivial(Var)).

% Flag evaluation

evaluate_flag_for_seq_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_seq_var(trivial(X),Value):-
        is_trivial_seq(X,Value).

evaluate_flag_for_ctx_var(Flag,Flag):-integer(Flag),!.
evaluate_flag_for_ctx_var(trivial(X),Value):-
        is_trivial_ctx(X,Value).
        
is_trivial_seq(eps,1):-!.
is_trivial_seq(seq([]),1):-!.
is_trivial_seq(S,0):-is_trivial_seq_aux(S,0),!.

is_trivial_seq_aux(seq([X|_]),0):-X\=seq_var(_),!.
is_trivial_seq_aux(seq([_|Y]),V):-is_trivial_seq_aux(seq(Y),V).

is_trivial_ctx(hole,1).
is_trivial_ctx(C,0):-is_trivial_ctx_aux(C,0),!.

is_trivial_ctx_aux([[X|L]|_],0):-is_trivial_ctx_aux([X|L],0),!.
is_trivial_ctx_aux([[_|_]|Tail],V):-!,is_trivial_ctx_aux(Tail,V).
is_trivial_ctx_aux([H|_],0):-H\=hole,H\=ctx_var(_),!.
is_trivial_ctx_aux([_|Tail],V):-is_trivial_ctx_aux(Tail,V).


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Selecting a branch in the 'sor' and 'cor' regular expression
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_branch(sor(R1,_R2),R1).
select_branch(sor(_R1,R2),R2).
select_branch(cor(R1,_R2),R1).
select_branch(cor(_R1,R2),R2).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
anonymous_to_singleton_variables_in_subst replaces in a substitution each
anonymous variable with a new temporary singleton variable
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

anonymous_to_singleton_variables_in_subst([],[]).
anonymous_to_singleton_variables_in_subst([X ---> RHS|Tail],[X--->RHS1|Tail1]):-
        anonymous_to_singleton_variables_in_term(RHS,RHS1),
        anonymous_to_singleton_variables_in_subst(Tail,Tail1).

anonymous_to_singleton_variables_in_term(c_,Var):-!,
        create_unique_temporary_ctxvar(ctx_var(c_An),Var).
anonymous_to_singleton_variables_in_term(s_,Var):-!,
        create_unique_temporary_seqvar(seq_var(s_An),Var).
anonymous_to_singleton_variables_in_term(f_,Var):-!,
        create_unique_temporary_funvar(fun_var(f_An),Var).
anonymous_to_singleton_variables_in_term(i_,Var):-!,
        create_unique_temporary_indvar(ind_var(i_An),Var).
anonymous_to_singleton_variables_in_term(seq(L),seq(L1)):-!,
        anonymous_to_singleton_variables_in_term(L,L1).
anonymous_to_singleton_variables_in_term([X],[Term]):-!,
        anonymous_to_singleton_variables_in_term(X,Term).
anonymous_to_singleton_variables_in_term([X|Tail],[X1|Tail1]):-!,
        anonymous_to_singleton_variables_in_term(X,X1),
        anonymous_to_singleton_variables_in_term(Tail,Tail1).
anonymous_to_singleton_variables_in_term(X,X).




/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
select_equations takes a matching problem and returns a list of all
matching equations that occur in the input, removing from it all regular
constraints.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

select_equations([],[]).
select_equations([X << Y|Tail], [X << Y |Tail1]):-!,
        select_equations(Tail,Tail1).
select_equations([_|Tail], L):-select_equations(Tail,L).



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purification
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

purify(Ein,Eout,NewConstraints):-
    purify(Ein,Eout,[],NewConstraints),
    !,
    NewConstraints\=[].


purify([],[],Acc,Acc):-!.

purify([sstar(T)|Tail1],[FreshSeqVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_seqvar(seq_var(s_),FreshSeqVar),!,
    purify(Tail1,Tail2,[[FreshSeqVar,sstar(T),0]|Acc],
           NewConstraints).

purify([sconc(T1,T2)|Tail1],[FreshSeqVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_seqvar(seq_var(s_),FreshSeqVar),!,
    purify(Tail1,Tail2,[[FreshSeqVar,sconc(T1,T2),0]|Acc],
           NewConstraints).

purify([sor(T1,T2)|Tail1],[FreshSeqVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_seqvar(seq_var(s_),FreshSeqVar),!,
    purify(Tail1,Tail2,[[FreshSeqVar,sor(T1,T2),0]|Acc],
           NewConstraints).

purify([cstar(C)|Tail1],[FreshCtxVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_ctxvar(ctx_var(c_),FreshCtxVar),!,
    purify(Tail1,Tail2,[[FreshCtxVar,cstar(C),0]|Acc],
           NewConstraints).

purify([cconc(C1,C2)|Tail1],[FreshCtxVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_ctxvar(ctx_var(c_),FreshCtxVar),!,
    purify(Tail1,Tail2,[[FreshCtxVar,cconc(C1,C2),0]|Acc],
           NewConstraints).

purify([cor(C1,C2)|Tail1],[FreshCtxVar|Tail2],Acc,NewConstraints):-
    create_unique_temporary_ctxvar(ctx_var(c_),FreshCtxVar),!,
    purify(Tail1,Tail2,[[FreshCtxVar,cor(C1,C2),0]|Acc],
           NewConstraints).

purify([[app,Ctx,Expr]|Tail1],[[FreshCtxVar,Expr1]|Tail2],Acc,NewConstraints):-
    create_unique_temporary_ctxvar(ctx_var(c_),FreshCtxVar),!,
        purify([Expr],[Expr1],[],Constr),
        append([[FreshCtxVar,Ctx,0]|Constr],Acc,NewAcc),
    purify(Tail1,Tail2,NewAcc,NewConstraints).

purify([[H|T]|Tail1],[[H1|T1]|Tail2],Acc,NewConstraints):-
        purify([H|T],[H1|T1],[],Constr),
    append(Constr,Acc,NewAcc),
    purify(Tail1,Tail2,NewAcc,NewConstraints).

purify([app,Ctx,Expr],[FreshCtxVar,Expr1],Acc,NewConstraints):-
    create_unique_temporary_ctxvar(ctx_var(c_),FreshCtxVar),!,
        purify([Expr],[Expr1],[[FreshCtxVar,Ctx,0]|Acc],NewConstraints).
     

purify([X|Tail1],[X|Tail2],Acc,NewConstraints):-
    purify(Tail1,Tail2,Acc,NewConstraints).








